#ifndef ANACAL_H
#define ANACAL_H

#include "anacal/image.h"
#include "anacal/fpfs.h"
#include "anacal/noise.h"
#include "anacal/psf.h"
#include "anacal/mask.h"
#include "anacal/ngmix.h"
#include "anacal/table.h"
#include "anacal/detector.h"
#include "anacal/task.h"

#endif // ANACAL_H
